# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Hamter CLI Nonbusiness Helper modules."""

import click_hotoffthehamster as click

from . import click_echo

__all__ = (
    'CrudeProgress',
)


class CrudeProgress(object):
    """"""
    def __init__(self, enabled):
        self.enabled = enabled

    # (lb): This is a very crude progress indicator.
    #   I should just Google and find one.
    #   But this does in a pinch. And really how useful is it?
    #   I'm working on a 400-factoid import file, and the only
    #   function that's noticeably slow is must_not_conflict_existing.
    def click_echo_current_task(self, task, no_clear=False):
        if not self.enabled:
            return

        def _click_echo_current_task():
            term_width = click.get_terminal_size()[0]
            cursor_to_leftmost_column()
            if not no_clear:
                click_echo(' ' * term_width, nl=False)  # "Clear" cursor line.
                cursor_to_leftmost_column()
            click_echo(task, nl=False)
            cursor_to_leftmost_column()
            # Move cursor past text.
            cursor_to_column_at(len(task) + 1)

        def cursor_to_leftmost_column():
            # FIXME: (lb): Can we use PPT to do cursoring? So that it detects terminal.
            #   Like, this'll work for me in my terminal, but what about, e.g., Windows?
            # MAGIC_CONTROL_CODE: Move cursor all the way left.
            click_echo(u"\u001b[1000D", nl=False)

        def cursor_to_column_at(col_num):
            # FIXME: (lb): Should be a PPT call or otherwise terminal-agnostic,
            #        and not specify a control code directly.
            click_echo(u"\u001b[" + str(col_num) + "C", nl=False)

        _click_echo_current_task()

    def start_crude_progressor(self, task_descrip):
        if not self.enabled:
            return

        self.click_echo_current_task(task_descrip)
        term_width = click.get_terminal_size()[0] - len(task_descrip) - 1
        dot_count = 0
        fact_sep = '.'
        return term_width, dot_count, fact_sep

    def step_crude_progressor(self, task_descrip, term_width, dot_count, fact_sep):
        if not self.enabled:
            return

        dot_count += 1
        if dot_count >= term_width:
            self.click_echo_current_task(task_descrip, no_clear=True)
            dot_count = 1
            fact_sep = ';' if fact_sep == '.' else '.'
        click_echo(fact_sep, nl=False)
        return term_width, dot_count, fact_sep




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import os

from gettext import gettext as _

from nark.helpers.app_dirs import NarkAppDirs

from ..termio import dob_in_user_warning

__all__ = (
    'AppDirs',
    'get_appdirs_subdir_file_path',
)


# ***
# *** `dob` AppDirs.
# ***

class DobAppDirs(NarkAppDirs):
    """Custom class that ensure appdirs exist."""
    def __init__(self, *args, **kwargs):
        """Add create flag value to instance."""
        super(DobAppDirs, self).__init__(*args, **kwargs)


AppDirs = DobAppDirs('dob')


# ***
# *** Shim function: get_appdirs_subdir_file_path.
# ***

DEFAULT_APPDIRS_FILE_BASENAME_FMT = '{}'


def get_appdirs_subdir_file_path(
    file_basename,
    dir_dirname,
    appdirs_dir=AppDirs.user_cache_dir,
    basename_fmt=DEFAULT_APPDIRS_FILE_BASENAME_FMT,
):
    """
    Return the path to a file stored in a subdirectory of an AppDirs directory.
    """
    subdir_path = os.path.join(appdirs_dir, dir_dirname)
    # (lb): So disrespectful! Totally accessing "hidden" method.
    AppDirs._ensure_directory_exists(subdir_path)
    full_path = os.path.join(subdir_path, basename_fmt.format(file_basename))
    if os.path.exists(full_path) and not os.path.isfile(full_path):
        msg = _(
            '{} At:\n  {}'
        ).format(
            _('UNEXPECTED: target path exists but not a file!'),
            full_path,
        )
        dob_in_user_warning(msg)
        full_path = None
    return full_path




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Module to provide u18n friendly help text strings for our controller needs."""

from gettext import gettext as _

from . import __arg0name__, __package_name__
from .config.fileboss import default_config_path
from .config.urable import ConfigUrable
from .termio import attr, bg, fg, highlight_value


__all__ = (
    'NEWBIE_HELP_WELCOME',
    'NEWBIE_HELP_ONBOARDING',
    'NEWBIE_HELP_CREATE_CONFIG',
    'NEWBIE_HELP_REPAIR_CONFIG',
    'NEWBIE_HELP_CREATE_STORE',
    # Private:
    #  'common_format',
    #  'section_heading',
)


def common_format():
    # (lb): Eh: This is a method because ENABLE_COLORS is False on startup,
    # and fg and attr will return empty strings if called when module is
    # sourced. So wait until help strings are built and it's known if color
    # or not.
    common_format = {
        'appname': highlight_value(__package_name__),
        'rawname': __package_name__,
        'codehi': fg('turquoise_2'),
        'reset': attr('reset'),
        'bold': attr('bold'),
        'italic': attr('italic'),
        'underlined': attr('underlined'),
        'wordhi': fg('chartreuse_3a'),
        'errcol': bg('red_1') + attr('bold'),
    }
    return common_format


# ***
# *** [INDUCTEE] help.
# ***

def NEWBIE_HELP_WELCOME(ctx):
    _help = _(
        """
        {color}┏━━━━━━━━━━━━━━━━━┓{reset}
        {color}┃ Welcome to dob! ┃{reset}
        {color}┗━━━━━━━━━━━━━━━━━┛{reset}
        """
    ).strip().format(
        color=(fg('spring_green_2a') + attr('bold')),
        # color=(fg('turquoise_4') + attr('bold')),
        # color=(fg('magenta_2a') + attr('bold')),
        # color=(fg('dark_orange_3b') + attr('bold')),
        **common_format(),
    )
    return _help


def section_heading(title):
    return (
        """
        {color}{title}{reset}
        {line_color}{sep:{sep}<{len_title}}{reset}
        """
    ).strip().format(
        title=title,
        # sep='-',
        sep='─',
        len_title=len(title),
        # color=fg('spring_green_2a'),
        # color=fg('dark_orange'),
        # color=fg('turquoise_4'),
        # line_color=fg('grey_78'),
        # line_color=fg('light_blue'),
        line_color='',
        color='',
        **common_format()
    )


def NEWBIE_HELP_ONBOARDING(ctx):
    # NOTE: This help is not automatically formatted like other text.
    _help = _(
        """
        {banner}

        Let’s get you setup!

        {init_title}
        To create a fresh, empty database, run:

          {cmd_color}{appname} init{reset}

        {upgrade_title}
        To learn how to import data from a previous version
        of dob, or to import from the old hamster app, run:

          {cmd_color}{appname} migrate -h{reset}

        {demo_title}
        If you’d like to get your hands dirty, you can demo
        the application with some example data that you can
        follow as a walk-through. Run:

          {cmd_color}{appname} demo{reset}
        \b
        """
    ).format(
        appname=__arg0name__,
        banner=NEWBIE_HELP_WELCOME(ctx),
        upgrade_title=section_heading(_('Import existing facts')),
        init_title=section_heading(_('Start from scratch')),
        demo_title=section_heading(_('Demo Dob')),
        help_title=section_heading(_('Such Help')),
        cmd_color=fg('spring_green_2a'),
        **common_format(),
    )
    return _help


def NEWBIE_HELP_CREATE_CONFIG(ctx, cfg_path):
    _help = _(
        """
        {errcol}ERROR: No config file found at: “{cfg_path}”{reset}

        Where's your config file??

        Verify and correct the configuration file path.

        The configuration file defaults to:

            {default_config_path}

        but you can override it using an environ:

            {envkey}=PATH

        or by specifying a global option:

            -C/--configfile PATH

        If you are certain the path is correct and you want to create
        a new configuration file at the path specified, run init, e.g.,:

            {rawname} -C "{cfg_path}" init
        """
    ).strip().format(
        # FIXME/2019-11-19 14:42: Make wrapper for format() with common colors defined.
        # - Maybe change errors to white on red, like here,
        #   but only for white on black terms (based on some setting?).
        cfg_path=cfg_path,
        default_config_path=default_config_path(),
        envkey=ConfigUrable.DOB_CONFIGFILE_ENVKEY,
        **common_format()
    )
    return _help


def NEWBIE_HELP_REPAIR_CONFIG(ctx, cfg_path):
    _help = _(
        """
        {errcol}ERROR: Please fix the config file found at: “{cfg_path}”{reset}

        The configuration file is at least missing the db.orm setting, if not others.

        Either edit and repair the confirguration file manually, or blow it away:

            {rawname} config create --force
        """
    ).strip().format(
        cfg_path=cfg_path,
        default_config_path=default_config_path(),
        envkey=ConfigUrable.DOB_CONFIGFILE_ENVKEY,
        **common_format()
    )
    return _help


def NEWBIE_HELP_CREATE_STORE(ctx, db_path, val_source):
    _help = _(
        """
        {errcol}ERROR: No database file found.{reset}

        - There was no file found at: {db_path}

        - The db.path value was set from the ‘{val_source}’ source.

        - For help on configuring settings, try the config help:
          \b
          {codehi}{rawname} config --help{reset}
        """
    ).strip().format(
        db_path=db_path,
        val_source=val_source,
        **common_format()
    )
    return _help




#!/usr/bin/env python

# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright

# Boilerplate documentation build configuration file,
# (Originally) created by sphinx-quickstart on Tue Jul 9 22:26:36 2013
# (and since somewhat modified to a make more palatable boilerplate).
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
import shlex
import sys
from pkg_resources import get_distribution

import sphinx_rtd_theme

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)

# ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
# ┃                                                                     ┃
# ┃ YOU/DEV: Customize this import and these strings for your project.  ┃

import dob_bright

project_dist = 'dob-bright'
package_name = 'dob-bright'
project_ghuser = 'hotoffthehamster'
project_ghrepo = project_dist
project_texinfo = 'One line description of project.'
project_docinfo = '{} Documentation'.format(project_dist)
project_htmlhelp_basename = 'DobConfigdoc'
project_copy = '2019-2020 Landon Bouma, HotOffThe Hamster, & contributors.'
project_auth = 'Landon Bouma'
project_orgn = 'HotOffThe Hamster'

exclude_patterns = [
    'CODE-OF-CONDUCT.rst',
    'CONTRIBUTING.rst',
    'README.rst',
]

# ┃                                                                     ┃
# ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
# Ref:
#   http://www.sphinx-doc.org/en/master/usage/extensions/index.html
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    # Google style docstrings
    # https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
    # https://google.github.io/styleguide/pyguide.html#38-comments-and-docstrings
    # https://google.github.io/styleguide/pyguide.html#383-functions-and-methods
    'sphinx.ext.napoleon',
    # For hyperlinks, e.g., :ref:`My Section Title`.
    'sphinx.ext.autosectionlabel',
]

# Prevent non local image warnings from showing.
suppress_warnings = ['image.nonlocal_uri']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = project_dist
copyright = project_copy
author = project_auth

# (lb): Using setuptools_scm magic, per
#   https://github.com/pypa/setuptools_scm#usage-from-sphinx
# we can call get_distribution, rather than hard coding herein.
#
# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution(project_dist).version
# The short X.Y version.
# - (lb): One place I see `release` used -- to name the browser page.
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to
# some non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built
# documents.
#keep_warnings = False

# -- Options for HTML output -------------------------------------------

# Ref:
#   http://www.sphinx-doc.org/en/master/usage/configuration.html#html-options
#   http://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# The theme to use for HTML and HTML Help pages.
# Ref:
#   https://sphinx-rtd-theme.readthedocs.io/en/latest/configuring.html
html_theme = 'sphinx_rtd_theme'

# 2020-03-29: There's a deprecation warning fixed upstream last year
# but the Sphinx package has not been released to PyPI since Feb, 2019.
# Here's the error:
#   writing additional pages...  search/<path>/.tox/docs/lib/python3.8/site-packages/
#       sphinx_rtd_theme/search.html:21: RemovedInSphinx30Warning: To modify script_files in
#       the theme is deprecated. Please insert a <script> tag directly in your theme instead.
html_theme_path = ["_themes", ]

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # Table of contents options.
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
    # Miscellaneous options.
    # 'canonical_url': '',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': False,
    'display_version': True,
    # prev_next_buttons_location: [bottom], top, both, or None.
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    # vcs_pageview_mode (upper-left navbar home button):
    #   With display_github: [blob], edit, or raw.
    #   #'vcs_pageview_mode': '',
    # style_nav_header_background: Default: '#2980B9'
    #   #'style_nav_header_background': '#2980B9',
}

# https://docs.readthedocs.io/en/latest/vcs.html?highlight=conf_py_path
html_context = {
    # Enable the "Edit in GitHub" link within the header of each page.
    'display_github': True,
    # Set the following variables to generate the resulting github URL for each page.
    # Format Template: https://{{ github_host|default("github.com") }}
    #   /{{ github_user }}/{{ github_repo }}/blob
    #   /{{ github_version }}{{ conf_py_path }}{{ pagename }}{{ suffix }}
    'github_user': project_ghuser,
    'github_repo': project_ghrepo,
    # Branch name.
    'github_version': 'develop/',
    # Path in the checkout to the docs root.
    'conf_py_path': 'docs/',
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as
# html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the
# top of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon
# of the docs.  This file should be a Windows icon file (.ico) being
# 16x16 or 32x32 pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# (lb): These work with alabaster, but are ignored by sphinx_rtd_theme.
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
        #'donate.html',
    ]
}

# Additional templates that should be rendered to pages, maps page names
# to template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer.
# Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer.
# Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages
# will contain a <link> tag referring to it.  The value of this option
# must be the base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = project_htmlhelp_basename

# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [(
    'index',
    '{}.tex'.format(package_name),
    project_docinfo,
    project_orgn,
    'manual',
), ]

# The name of an image file (relative to this directory) to place at
# the top of the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings
# are parts, not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(
    'index',
    package_name,
    project_docinfo,
    [project_orgn],
    1,
), ]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [(
    'index',
    package_name,
    project_docinfo,
    project_orgn,
    package_name,
    project_texinfo,
    'Miscellaneous',
), ]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Methods for common terminal echo operations."""

import click_hotoffthehamster as click

from .paging import click_echo
from .style import attr, fg

__all__ = (
    # Callers might want to import click_echo from this module, because it
    # feels more natural here (but its in paging module because pager-aware).
    'click_echo',
    'echo_block_header',
    'highlight_value',
    # PRIVATE:
    # '__format_block_header',
)


# ***

def echo_block_header(title, **kwargs):
    click_echo()
    click_echo(__format_block_header(title, **kwargs))


def __format_block_header(title, sep='━', full_width=False):
    """"""
    def _fact_block_header():
        header = []
        append_highlighted(header, title)
        append_highlighted(header, hr_rule())
        return '\n'.join(header)

    def append_highlighted(header, text):
        highlight_col = 'red_1'
        header.append('{}{}{}'.format(
            fg(highlight_col),
            text,
            attr('reset'),
        ))

    def hr_rule():
        if not full_width:
            horiz_rule = sep * len(title)
        else:
            # NOTE: When piping (i.e., no tty), width defaults to 80.
            term_width = click.get_terminal_size()[0]
            horiz_rule = '─' * term_width
        return horiz_rule

    return _fact_block_header()


# ***

def highlight_value(msg):
    highlight_color = 'medium_spring_green'
    return '{}{}{}'.format(fg(highlight_color), msg, attr('reset'))




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import os

from gettext import gettext as _

from configobj import ConfigObj, DuplicateError

from ..termio import dob_in_user_exit

from .app_dirs import AppDirs

__all__ = (
    'default_config_path',
    'empty_config_obj',
    'load_config_obj',
    'write_config_obj',
)


# ***

def default_config_path():
    config_dir = AppDirs.user_config_dir
    config_filename = 'dob.conf'
    configfile_path = os.path.join(config_dir, config_filename)
    return configfile_path


# ***

def empty_config_obj(configfile_path):
    config_obj = ConfigObj(
        configfile_path,
        write_empty_values=False,
        # Note that ConfigObj has a raise_errors param, but if False, it
        # just defers the error, if any; it'll still get raised, just at
        # the end. So what's the point? -(lb)
        #   raise_errors=False,
    )
    return config_obj


# ***

def load_config_obj(configfile_path):
    """"""

    def _load_config_obj():
        try:
            config_obj = empty_config_obj(configfile_path)
        except DuplicateError as err:
            # (lb): The original (builtin) configparser would let you
            # choose to error or not on duplicates, but the ConfigObj
            # library (which is awesome in many ways) does not have
            # such a feature (it's got a raise_errors that does not
            # do the trick). Consequently, unless we code a way around
            # this, we gotta die on duplicates. Sorry, User! Seems
            # pretty lame. But also seems pretty unlikely.
            exit_duplicates(str(err))

        return config_obj

    def exit_duplicates(err):
        msg = _(
            'ERROR: Your config file at “{}” has a duplicate setting: “{}”'
        ).format(configfile_path, str(err))
        dob_in_user_exit(msg)

    return _load_config_obj()


# ***

def write_config_obj(config_obj):
    def _write_config_obj():
        if not config_obj.filename:
            raise AttributeError('ConfigObj missing ‘filename’')
        ensure_dirs(config_obj.filename)
        try:
            config_obj.write()
        except UnicodeEncodeError as err:
            die_write_failed(config_obj, err)

    def ensure_dirs(filename):
        # Avoid: FileNotFoundError: [Errno 2] No such file or directory: ....
        configfile_dir = os.path.dirname(filename)
        if configfile_dir and not os.path.lexists(configfile_dir):
            os.makedirs(configfile_dir)

    def die_write_failed(config_obj, err):
        # E.g.,:
        #   UnicodeEncodeError: 'ascii' codec can't encode character
        #     '\u2018' in position 1135: ordinal not in range(128)
        msg = _(
            'ERROR: Failed to write file at “{}”: “{}”\n'
            'Perhaps unknown character(s): {}'
        ).format(
            config_obj.filename,
            str(err),
            err.object[err.start:err.end],
        )
        dob_in_user_exit(msg)

    return _write_config_obj()




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""``hamster``, ``hamster``, ``hamster``!!! a cuddly, furry time tracker."""

import os
import sys

__all__ = (
    '__arg0name__',
    '__author__',
    '__author_email__',
    '__package_name__',
    '__resolve_vers__',
)

# (lb): These are duplicated in setup.cfg:[metadata], but not sure how to DRY.
#   Fortunately, they're not likely to change.
__author__ = 'HotOffThe Hamster'
__author_email__ = 'hotoffthehamster+dob@gmail.com'

# (lb): Not sure if the package name is available at runtime. Seems kinda meta,
# anyway, like, Who am I? I just want to avoid hard coding this string in docs.
__package_name__ = 'dob'
__arg0name__ = os.path.basename(sys.argv[0])


def __resolve_vers__():
    """Returns the installed package version, or '<none>'.

    In lieu of always setting __version__ -- and always loading pkg_resources --
    use a method to avoid incurring startup costs if the version is not needed.
    """
    try:
        import setuptools_scm
        # For whatever reason, relative_to does not work, (lb) thought it would.
        #   return setuptools_scm.get_version(relative_to=__file__)
        pkg_dirname = os.path.dirname(os.path.dirname(__file__))
        # See if parent directory (of this file) contains .git/.
        proj_git = os.path.join(pkg_dirname, '.git')
        if os.path.exists(proj_git):
            # Get version from setuptools_scm, and git tags.
            # This is similar to a developer running, e.g.,
            #   python setup.py --version
            return setuptools_scm.get_version(root=pkg_dirname)
    except ImportError:
        # ModuleNotFoundError added py3.6. Until then, ImportError.
        from pkg_resources import get_distribution, DistributionNotFound
        try:
            return get_distribution(__package_name__).version
        except DistributionNotFound:
            return '<none>'
    else:
        return '<none>'




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import json
import os

from gettext import gettext as _

from config_decorator import ConfigDecorator, KeyChainedValue

from nark.config import ConfigRoot
from nark.config.log_levels import get_log_level_safe, must_verify_log_level
from nark.helpers.parsing import FACT_METADATA_SEPARATORS

from .app_dirs import AppDirs

__all__ = (
    'decorate_config',
    'DobConfigurableDev',
    'DobConfigurableEditor',
    'DobConfigurableFact',
    'DobConfigurableLog',
    'DobConfigurableTerm',
    # PRIVATE:
    # '_styling_file_path',
)


KeyChainedValue._envvar_prefix = 'DOB_'


# ***
# *** Client (dob) Config.
# ***

def _styling_file_path(basename):
    dirname = AppDirs.user_config_dir
    subdir = 'styling'
    # E.g., /home/user/config/.dob/styling/{basename}
    return os.path.join(dirname, subdir, basename)


@ConfigRoot.section('editor')
class DobConfigurableEditor(object):
    """"""

    def __init__(self, *args, **kwargs):
        # See comment in NarkConfigurable:
        #   Cannot super because decorator shenanigans.
        # DEATH: super(DobConfigurableEditor, self).__init__(*args, **kwargs)
        # FIXME/2019-11-30 17:47: Could test again: (lb): I might have fixed issue.
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, center the interface in the terminal (on `dob edit`)"),
    )
    def centered(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        # FIXME/2019-11-29: (lb): Help Docs: Add command to list dob + pygments.lexers.
        _("dob or Pygments lexer. See: https://pygments.org/docs/lexers/"),
    )
    def lexer(self):
        # (lb): I'm a reSTie, personally, so let's just default to that.
        # (Not to dishonor other markups, I'm friends with markdownies,
        # I even LaTeXeD once in my life, and let's not talk about that
        # one time I was into this really weird markup called Textile.)
        # MAYBE/2019-11-29: (lb): Perhaps default to '' (and move this to *my* config).
        return 'RstLexer'

    # ***

    @property
    @ConfigRoot.setting(
        _("UX style: '' or 'default' applies none;"
          "'night', 'light', or 'color' applies builtin;"
          " or name your own (see: styles_fpath)."),
    )
    def styling(self):
        # Do not specify a style (or specify 'default') so that the UX is not
        # styled, but uses the current terminal color scheme.
        return ''

    # ***

    @property
    @ConfigRoot.setting(
        # FIXME/2019-11-30: Backlog entry: how best to split long help string.
        # - One option: Split here with explicit newlines. Works well for:
        #                 dob config dump -T tabulate
        #               but does not work with -T friendly
        #               - We could default to tabulate...
        #                 or maybe help() callback that customizes
        #                 text based on -T option?
        # - Another option: Truncate after first period '.' on global dob-bright-dump,
        #                    but include complete help on dob-bright-dump-setting (e.g.,)
        _("Path to file containing newline-separated ‘act@gory’ and ‘#tag’\n"
          "names to omit from completions and suggestions (regex-aware)."),
        name='ignore_fpath',
    )
    def ignore_list_path(self):
        # E.g., /home/user/config/.dob/styling/ignore.list
        return _styling_file_path(basename='ignore.list')

    # ***

    @property
    @ConfigRoot.setting(
        _("Path to file containing UX styles used by “editor.styling”."),
        name='styles_fpath',
    )
    def styles_conf_path(self):
        # E.g., /home/user/config/.dob/styling/styles.conf
        return _styling_file_path(basename='styles.conf')

    # ***

    @property
    @ConfigRoot.setting(
        _("Path to file defining object matching used to stylize UX."),
        name='stylit_fpath',
    )
    def stylit_conf_path(self):
        # E.g., /home/user/config/.dob/styling/queries.conf
        return _styling_file_path(basename='stylit.conf')

    # ***


# ***

@ConfigRoot.section('fact')
class DobConfigurableFact(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("Acceptable separator(s) to delimit Fact meta data from description."),
    )
    def separators(self):
        # Rather than `return ''` and act like there are no separators, show
        # the default value that nark uses, so that user are better educated.
        # FYI: json.dumps([",", ":"]) → '[",", ":"]'
        return json.dumps(FACT_METADATA_SEPARATORS)


# ***

@ConfigRoot.section('dev')
class DobConfigurableDev(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("The log level for frontend (dob) squaller"
          " (using Python logging library levels)"),
        validate=must_verify_log_level,
        conform=get_log_level_safe,
    )
    def cli_log_level(self):
        return 'WARNING'

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, enables features for developing dob"
          " (e.g., stop at REPL on affirm faults)."),
    )
    def catch_errors(self):
        return False


# ***

@ConfigRoot.section('log')
class DobConfigurableLog(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    # MAYBE/2019-11-29: (lb): I'm not sure utility of setting logfile name,
    # but not the path. Perhaps add log.dirname option.
    # - Or
    @property
    @ConfigRoot.setting(
        # MAYBE/2019-11-29: Resolve appdirs path -- or somehow show user full log path.
        # Or, if (hypothetical) log.dirname option added, substitute in help text here.
        _("Filename of dob log under AppDirs.user_log_dir"),
    )
    def filename(self):
        return 'dob.log'

    # ***

    # User should not set logfile_path in config, because it's
    # generated, but we make it a pseudo-setting so the code can
    # access just the same as other settings (and the value depends
    # on another config value, the log_filename, so it sorta is a
    # config value, just a derived one).
    @property
    @ConfigRoot.setting(
        _('Generated value.'),
        ephemeral=True,
    )
    def filepath(self):
        if self is None:
            # If called on class definition, before kcv._section set, return
            # empty string, which indicates the value type. However, because
            # ephemeral=True, _deduce_value_type won't call this method with
            # self set to None -- meaning, this return statement technically
            # unreachable. But we'll still keep it, for completeness.
            return ''
        log_dir = AppDirs.user_log_dir
        # Note that self is the root ConfigDecorator, not the DobConfigurableLog.
        log_filename = self['filename']
        return os.path.join(log_dir, log_filename)

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, use color in log files (useful if tailing logs in terminal)"),
    )
    def use_color(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, send logs to the console, not the file"),
    )
    def use_console(self):
        return True


# ***

@ConfigRoot.section('term')
class DobConfigurableTerm(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("Path to directory where export files are created"),

        # FIXME/2019-11-17 00:56: Nothing uses export_path.
        #   So hidden for now. Delete later, or implement.
        hidden=True,
    )
    def export_path(self):
        """
        Return path to save exports to.
        File extension will be added by export method.
        """
        # FIXME: Not sure who this'll be used... maybe two separate config
        #        values, one editable and one generated, like log_filename
        #        and logfile_path. But first, we need a feature that exports!
        #        Until then... either keep this setting hidden, or delete it!
        return os.path.join(AppDirs.user_data_dir, 'export')

    # ***

    @property
    @ConfigRoot.setting(
        _("The filename suffix to tell EDITOR so it can determine highlighting"),
    )
    def editor_suffix(self):
        return ''

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, show copyright greeting before every command."),
    )
    def show_greeting(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, use color and font ornamentation in output."),
    )
    def use_color(self):
        return True

    # ***

    @property
    @ConfigRoot.setting(
        _("If True, send application output to terminal pager."),
    )
    def use_pager(self):
        return False


# ***

def decorate_config(config):
    """Wraps or ensures the supplied config dict is a nark config decorator.

    Note: This is mostly for the ``nark/tests``, as the dob client will
    explicitly prepare a decorated config and pass that to nark. But to
    make it easier to write fixtures for ``nark/tests``, this method
    accepts a dictionary and returns a decorated config.
    """
    if isinstance(config, ConfigDecorator):
        return config
    # Caution: Remember that ConfigRoot is a weird Singleton.
    # FIXME/2020-01-07: (lb): Can probably put ConfigRoot et al in a getter
    #                         so that it doesn't have to be a module global.
    config_root = ConfigRoot
    config_root.forget_config_values()
    config_root.update(config)
    return config_root




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import pytest

from nark.helpers import logging as logging_helpers

from dob_bright.config import decorate_config, app_dirs, fileboss
from dob_bright.config.fileboss import write_config_obj
from dob_bright.config.urable import ConfigUrable


class TestSetupLogging(object):
    """Make sure that our logging setup is executed as expected."""

    def test_setup_logging_and_log_level(self, controller):
        """
        Test that library and client logger have log level set according to config.
        """
        controller.setup_logging()
        assert controller.lib_logger.level == (
            logging_helpers.resolve_log_level(
                controller.config['dev.lib_log_level']
            )[0]
        )
        assert controller.client_logger.level == (
            logging_helpers.resolve_log_level(
                controller.config['dev.cli_log_level']
            )[0]
        )

    def test_setup_logging_log_console_true(self, controller):
        """Ensure if console logging, lib and client have streamhandlers."""
        controller.config['log.use_console'] = True
        controller.setup_logging()
        assert isinstance(
            controller.client_logger.handlers[0],
            logging.StreamHandler,
        )
        assert isinstance(
            controller.client_logger.handlers[1],
            logging.FileHandler,
        )
        assert isinstance(
            controller.lib_logger.handlers[0],
            logging.StreamHandler,
        )
        assert isinstance(
            controller.lib_logger.handlers[1],
            logging.FileHandler,
        )
        assert len(controller.client_logger.handlers) == 2
        assert len(controller.lib_logger.handlers) == 2
        assert controller.client_logger.handlers[0].formatter

    def test_setup_logging_no_logging(self, controller):
        """Make sure that if no logging enabled, our loggers don't have any handlers."""
        controller.setup_logging()
        # Default loggers are set up in ~/.cache/<app>/log/<app>.log
        assert len(controller.lib_logger.handlers) == 1
        assert len(controller.client_logger.handlers) == 1

    def test_setup_logging_log_file_true(self, controller, appdirs):
        """
        Make sure that if we enable logfile_path, both loggers receive ``FileHandler``.
        """
        controller.config['log.filepath'] = os.path.join(
            appdirs.user_log_dir, 'foobar.log',
        )
        controller.setup_logging()
        assert isinstance(
            controller.lib_logger.handlers[0],
            logging.FileHandler,
        )
        assert isinstance(
            controller.client_logger.handlers[0],
            logging.FileHandler,
        )


class TestGetConfig(object):
    """Make sure that turning a config instance into proper config dictionaries works."""

    @pytest.mark.parametrize('cli_log_level', ['debug'])
    def test_log_levels_valid(self, cli_log_level, config_instance):
        """
        Make sure *string loglevels* translates to their respective integers properly.
        """
        config_obj = config_instance(cli_log_level=cli_log_level)
        assert config_obj['dev']['cli_log_level'] == cli_log_level
        config = decorate_config(config_obj)
        assert config['dev']['cli_log_level'] == 10
        assert config['dev.cli_log_level'] == 10
        assert config.asobj.dev.cli_log_level.value == 10

    @pytest.mark.parametrize('cli_log_level', ['foobar'])
    def test_log_levels_invalid(self, cli_log_level, config_instance, capsys):
        """Test that invalid *string loglevels* raise ``ValueError``."""
        config_obj = config_instance(cli_log_level=cli_log_level)
        with pytest.raises(
            ValueError,
            match=r"^Unrecognized value for setting ‘cli_log_level’: “foobar”.*"
        ):
            _config = decorate_config(config_obj)  # noqa: F841 unused local
        out, err = capsys.readouterr()
        assert out == ''
        assert err == ''

    def test_invalid_store(self, config_instance):
        """Make sure that passing an ORM other than 'sqlalchemy' raises an exception."""
        config_obj = config_instance(orm='foobar')
        match_former = r'Unrecognized value for setting ‘orm’'
        match_latter = r'“foobar” \(Choose from: ‘sqlalchemy’\)'
        with pytest.raises(
            ValueError, match=r"^{}: {}$".format(match_former, match_latter),
        ):
            _config = decorate_config(config_obj)  # noqa: F841 unused local

    def test_non_sqlite(self, config_instance):
        """Make sure that passing a postgres config works.

        Albeit actual postgres connections not tested."""
        confnstnc = config_instance(engine='postgres')
        config = decorate_config(confnstnc)
        assert config['db.host'] == confnstnc['db']['host']
        assert config['db.port'] == confnstnc['db']['port']
        assert config['db.name'] == confnstnc['db']['name']
        assert config['db.user'] == confnstnc['db']['user']
        assert config['db.password'] == confnstnc['db']['password']


class TestGetConfigInstance(object):
    def test_no_file_present(self, appdirs, mocker):
        # In lieu of testing from completely vanilla account, ensure config file does
        # not exist (which probably exists for your user at ~/.config/dob/dob.conf).
        # NOTE: AppDirs is a module-scope object with immutable attributes, so we
        # need to mock the entire object (i.e., cannot just patch attribute itself).
        app_dirs_mock = mock.Mock()
        app_dirs_mock.configure_mock(user_config_dir='/XXX')
        app_dirs_mock.configure_mock(user_data_dir='/XXX')
        mocker.patch.object(fileboss, 'AppDirs', app_dirs_mock)
        self.configurable = ConfigUrable()
        self.configurable.load_config(configfile_path=None)
        assert len(list(self.configurable.config_root.items())) > 0
        assert self.configurable.cfgfile_exists is False

    def test_file_present(self, config_instance):
        """Make sure we try parsing a found config file."""
        self.configurable = ConfigUrable()
        self.configurable.load_config(configfile_path=None)
        cfg_val = self.configurable.config_root['db']['orm']
        assert cfg_val == config_instance()['db']['orm']
        assert config_instance() is not self.configurable.config_root

    def test_config_path_getter(self, appdirs, mocker):
        """Make sure the config target path is constructed to our expectations."""
        mocker.patch('dob_bright.config.fileboss.load_config_obj')
        # DRY?/2020-01-09: (lb): Perhaps move repeated ConfigUrable code to fixture.
        self.configurable = ConfigUrable()
        self.configurable.load_config(configfile_path=None)
        # 'dob.conf' defined and used in dob_bright.config.fileboss.default_config_path.
        expectation = os.path.join(appdirs.user_config_dir, 'dob.conf')
        assert fileboss.load_config_obj.called_with(expectation)


class TestWriteConfigFile(object):
    def test_file_is_written(self, filepath, config_instance):
        """Ensure file is written. Content not checked; that's ConfigObj's job."""
        config_obj = config_instance()
        write_config_obj(config_obj)
        assert os.path.lexists(config_obj.filename)

    def test_non_existing_path(self, tmpdir, filename, config_instance):
        """Make sure that the path-parents are created if not present."""
        filepath = os.path.join(tmpdir.strpath, filename)
        assert os.path.lexists(filepath) is False
        config_obj = config_instance()
        config_obj.filename = filepath
        write_config_obj(config_obj)
        assert os.path.lexists(config_obj.filename)


class TestDobAppDirs(object):
    """AppDirs tests."""

    def _test_app_dir_returns_directoy(self, app_dirname, tmpdir, **kwargs):
        """Make sure method returns directory."""
        path = tmpdir.strpath
        with mock.patch(
            'appdirs.{}'.format(app_dirname),
            new_callable=mock.PropertyMock,
        ) as mock_app_dir:
            mock_app_dir.return_value = path
            appdir = app_dirs.DobAppDirs('dob')
            assert getattr(appdir, app_dirname) == path
            # (lb): Guh. After py3.5 dropped, we could simplify this to:
            #   mock_app_dir.assert_called_once()
            # Until then, gotta specify args and kwargs.
            # MAYBE/2020-01-29: Ha! Dropped py3.5 because PTK3 dropped it!
            # - So now we could see about simplifying this.
            kwargs['version'] = None
            mock_app_dir.assert_called_once_with('dob', None, **kwargs)

    def _test_app_dir_creates_file(self, app_dirname, create, tmpdir, faker, **kwargs):
        """Make sure that path creation depends on ``create`` attribute."""
        path = os.path.join(tmpdir.strpath, '{}/'.format(faker.word()))
        # We want NarkAppDirs's call to appdirs.XXX_dir to return our /tmp path.
        # (lb): Note that mocker (from pytest-mock) merely takes care of teardown,
        #   which is also accomplished with your simply use with-mock, as follows.
        with mock.patch(
            'appdirs.{}'.format(app_dirname),
            new_callable=mock.PropertyMock,
            return_value=path,
        ) as mock_app_dir:
            appdir = app_dirs.DobAppDirs('dob')
            appdir.create = create
            # DEVS: Weird: If this assert fires and you're running `py.test --pdb`,
            # entering e.g., `appdir.user_data_dir` at the pdb prompt shows the
            # non-mocked value! But if you capture the value first and print it,
            # it's correct! So in code you'd have:
            #   show_actual = appdir.user_data_dir
            # And in pdb you'd type:
            #   (pdb) show_actual
            #   '/tmp/pytest-of-user/pytest-1142/test_user_data_dir_creates_fil0/relationship/'
            #   (pdb) appdir.user_data_dir
            #   '/home/user/.local/share/dob'
            assert os.path.exists(getattr(appdir, app_dirname)) is create
            # MAYBE/2020-01-29: New min. Py now 3.6, so might be able to simplify:
            #   mock_app_dir.assert_called_once()
            # but currently doing it hard way _with() args and kwargs (circa py3.5).
            kwargs['version'] = None
            mock_app_dir.assert_called_once_with('dob', None, **kwargs)

    # ***

    def test_user_data_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy(
            'user_data_dir', tmpdir, roaming=False,
        )

    @pytest.mark.parametrize('create', [True, False])
    def test_user_data_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file(
            'user_data_dir', create, tmpdir, faker, roaming=False,
        )

    # ---

    def test_site_data_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy(
            'site_data_dir', tmpdir, multipath=False,
        )

    @pytest.mark.parametrize('create', [True, False])
    def test_site_data_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file(
            'site_data_dir', create, tmpdir, faker, multipath=False,
        )

    # ---

    def test_user_config_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy(
            'user_config_dir', tmpdir, roaming=False,
        )

    @pytest.mark.parametrize('create', [True, False])
    def test_user_config_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file(
            'user_config_dir', create, tmpdir, faker, roaming=False,
        )

    # ---

    def test_site_config_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy(
            'site_config_dir', tmpdir, multipath=False,
        )

    @pytest.mark.parametrize('create', [True, False])
    def test_site_config_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file(
            'site_config_dir', create, tmpdir, faker, multipath=False,
        )

    # ---

    def test_user_cache_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy('user_cache_dir', tmpdir)

    @pytest.mark.parametrize('create', [True, False])
    def test_user_cache_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file('user_cache_dir', create, tmpdir, faker)

    # ---

    def test_user_log_dir_returns_directoy(self, tmpdir):
        """Make sure method returns directory."""
        self._test_app_dir_returns_directoy('user_log_dir', tmpdir)

    @pytest.mark.parametrize('create', [True, False])
    def test_user_log_dir_creates_file(self, tmpdir, faker, create):
        """Make sure that path creation depends on ``create`` attribute."""
        self._test_app_dir_creates_file('user_log_dir', create, tmpdir, faker)




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma, © 2015-2016 Eric Goller. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""
Fixtures available to the tests/.

- In general, fixtures should return a single instance.

- If a fixture is a factory, its name should reflect that.

- A fixture that is parametrized should be suffixed with
  ``_parametrized`` to imply it has increased complexity.
"""

import codecs
import os
import pytest
from configobj import ConfigObj

import fauxfactory
from pytest_factoryboy import register

from nark.config import decorate_config
from nark.items.fact import Fact

from dob_bright.config import app_dirs  # Needs container of AppDirs.
from dob_bright.controller import Controller

from nark.tests import factories

register(factories.CategoryFactory)
register(factories.ActivityFactory)
register(factories.FactFactory)

test_lib_log_level = 'WARNING'
test_cli_log_level = 'WARNING'
# DEV: Uncomment to see more log trace while testing:
test_lib_log_level = 'DEBUG'
test_cli_log_level = 'DEBUG'


# ***

@pytest.fixture
def filename():
    """Provide a filename string."""
    return fauxfactory.gen_utf8()


@pytest.fixture
def filepath(tmpdir, filename):
    """Provide a fully qualified pathame within our tmp-dir."""
    return os.path.join(tmpdir.strpath, filename)


# ***

@pytest.fixture
def appdirs(mocker, tmpdir):
    """Provide mocked version specific user dirs using a tmpdir."""
    def ensure_directory_exists(directory):
        if not os.path.lexists(directory):
            os.makedirs(directory)
        return directory

    mock_dirs = mocker.MagicMock()
    mock_dirs.user_config_dir = ensure_directory_exists(
        os.path.join(tmpdir.mkdir('config').strpath, 'dob/'),
    )
    mock_dirs.user_data_dir = ensure_directory_exists(
        os.path.join(tmpdir.mkdir('data').strpath, 'dob/'),
    )
    mock_dirs.user_cache_dir = ensure_directory_exists(
        os.path.join(tmpdir.mkdir('cache').strpath, 'dob/'),
    )
    mock_dirs.user_log_dir = ensure_directory_exists(
        os.path.join(tmpdir.mkdir('log').strpath, 'dob/'),
    )

    app_dirs.AppDirs = mock_dirs

    return mock_dirs


# ***

@pytest.fixture
def config_root(nark_config, dob_config):
    """Provide a generic baseline configuration."""
    config_root = decorate_config(nark_config)
    config_root.update(dob_config)
    # Beware config_decorator settings' _conform_f(), which mutates source values.
    # To layer config collections, be sure to use the original, pre-mutation values.
    config = config_root.as_dict(unmutated=True)
    return config


# This method essentially same as: nark:tests/conftest.py::base_config.
@pytest.fixture
def nark_config(tmpdir):
    """
    Provide a static backend config fixture.
    """
    # FETREQ/2020-01-09: (lb): Support dot-notation in dict keys on `update`.
    # - For now, create deep dictionary; not flat with dotted key names.
    return {
        'db': {
            'orm': 'sqlalchemy',
            'engine': 'sqlite',
            'path': ':memory:',
            # MAYBE/2019-02-20: (lb): Support for alt. DBMS is wired, but not tested.
            #   'host': '',
            #   'port': '',
            #   'name': '',
            #   'user': '',
            #   'password': '',
        },
        'dev': {
            'lib_log_level': test_lib_log_level,
            'sql_log_level': 'WARNING',
        },
        'time': {
            # 2019-02-20: (lb): Note that allow_momentaneous=False probably Bad Idea,
            #                   especially for user upgrading from legacy hamster db.
            'allow_momentaneous': True,

            # MAYBE/2019-02-20: (lb): I don't day_start, so probably broke; needs tests.
            #   'day_start': datetime.time(hour=0, minute=0, second=0).isoformat(),
            #   'day_start': datetime.time(hour=5, minute=0, second=0).isoformat(),
            'day_start': '',

            # MAYBE/2019-02-20: (lb): Perhaps test min-delta, another feature I !use!
            #   'fact_min_delta': '60',
            'fact_min_delta': '0',

            # FIXME/2019-02-20: (lb): Implement tzawareness/tz_aware/timezone sanity.
            'tz_aware': False,

            # FIXME/2019-02-20: (lb): Needs testing, e.g.,
            #   'default_tzinfo': 'America/Menominee',
            'default_tzinfo': '',
        },
    }


@pytest.fixture
def dob_config(tmpdir):
    """
    Provide a static client config fixture.
    """
    return {
        # 'editor.centered': '',
        'editor.centered': 'True',

        'editor.lexer': '',

        # Devmode catch_errors could be deadly under test, as it sets a trace trap.
        'dev.catch_errors': False,

        'term.editor_suffix': '',

        # The default export path is '', i.e., local directory. Use /tmp instead.
        # 'term.export_path': '',  # Default.
        'term.export_path': os.path.join(tmpdir.mkdir('export').strpath, 'export'),

        # Disable color, otherwise tests will have to look for color codes.
        'log.use_color': False,

        # Don't log to console, otherwise tests have to deal with that noise.
        # 'log_console': True,  # Default.
        'log.use_console': False,

        # The default log filename does not need to be changed.
        # 'log_filename': 'dob.log',  # Default.
        # See also:
        #  'logfile_path': '',  # Generated value.

        'dev.cli_log_level': test_cli_log_level,

        'fact.separators': '',  # [,:\n]

        'term.show_greeting': False,

        'editor.styling': '',

        'term.use_color': False,
        'term.use_pager': False,
    }


@pytest.fixture
def config_instance(tmpdir, faker):
    """Provide a (dynamicly generated) ConfigObj instance."""

    def generate_config(**kwargs):
        cfg_dict = generate_dict(**kwargs)
        # NOPE: You'd overwrite your user's file with the default path:
        #   from dob_bright.config.fileboss import default_config_path
        #   configfile_path = default_config_path()
        configfile_path = os.path.join(tmpdir, 'dob.conf')
        config = ConfigObj(configfile_path)
        config.merge(cfg_dict)
        return config

    def generate_dict(**kwargs):
        cfg_dict = {}

        # ***

        cfg_db = {}
        cfg_dict['db'] = cfg_db

        cfg_db.setdefault('orm', kwargs.get('orm', 'sqlalchemy'))
        cfg_db.setdefault('engine', kwargs.get('engine', 'sqlite'))
        # HARDCODED: This filename value does not matter, really.
        db_path = os.path.join(tmpdir.strpath, 'dob.sqlite')
        cfg_db.setdefault('path', kwargs.get('path', db_path))
        cfg_db.setdefault('host', kwargs.get('host', ''))
        cfg_db.setdefault('port', kwargs.get('port', ''))
        cfg_db.setdefault('name', kwargs.get('name', ''))
        cfg_db.setdefault('user', kwargs.get('user', '')),
        cfg_db.setdefault('password', kwargs.get('password', ''))

        # ***

        cfg_dev = {}
        cfg_dict['dev'] = cfg_dev

        lib_log_level = kwargs.get('lib_log_level', test_lib_log_level)
        cfg_dev.setdefault('lib_log_level', lib_log_level)
        sql_log_level = kwargs.get('sql_log_level', 'WARNING')
        cfg_dev.setdefault('sql_log_level', sql_log_level)

        # ***

        cfg_time = {}
        cfg_dict['time'] = cfg_time

        # (lb): Need to always support momentaneous, because legacy data bugs.
        # cfg_time.setdefault('allow_momentaneous', 'False')
        cfg_time.setdefault('allow_momentaneous', 'True')

        # day_start = kwargs.get('day_start', '')
        day_start = kwargs.get('day_start', '00:00:00')
        cfg_time.setdefault('day_start', day_start)

        # fact_min_delta = kwargs.get('fact_min_delta', '0')
        fact_min_delta = kwargs.get('fact_min_delta', '60')
        cfg_time.setdefault('fact_min_delta', fact_min_delta)

        cfg_time.setdefault('tz_aware', kwargs.get('tz_aware', 'False'))
        # FIXME/2019-02-20: (lb): Fix timezones. And parameterize, e.g.,
        #  default_tzinfo = kwargs.get('default_tzinfo', 'America/Menominee')
        default_tzinfo = kwargs.get('default_tzinfo', '')
        cfg_time.setdefault('default_tzinfo', default_tzinfo)

        # ***

        cfg_editor = {}
        cfg_dict['editor'] = cfg_editor

        cfg_editor.setdefault('centered', False)
        cfg_editor.setdefault('lexer', '')
        cfg_editor.setdefault('styling', '')

        # ***

        cfg_fact = {}
        cfg_dict['fact'] = cfg_fact

        cfg_fact.setdefault('separators', '')  # [,:\n]

        # ***

        assert('dev' in cfg_dict)

        cli_log_level = kwargs.get('cli_log_level', test_cli_log_level)
        cfg_dev.setdefault('cli_log_level', cli_log_level)
        cfg_dev.setdefault('catch_errors', 'False')

        # ***

        cfg_log = {}
        cfg_dict['log'] = cfg_log

        cfg_log.setdefault('filename', kwargs.get('log_filename', faker.file_name()))
        # The log_filename is used to make log.filepath, which we don't need to set.
        cfg_log.setdefault('use_color', 'False')
        cfg_log.setdefault('use_console', kwargs.get('log_console', 'False'))

        # ***

        cfg_term = {}
        cfg_dict['term'] = cfg_term

        cfg_term.setdefault('export_path', '')
        cfg_term.setdefault('editor_suffix', '')
        cfg_term.setdefault('show_greeting', 'False')
        cfg_term.setdefault('use_color', 'True')
        cfg_term.setdefault('use_pager', 'False')

        # ***

        return cfg_dict

    return generate_config


@pytest.fixture
def config_file(config_instance, appdirs):
    """Provide a config file store under our fake config dir."""
    conf_path = os.path.join(appdirs.user_config_dir, 'dob.conf')
    with codecs.open(conf_path, 'w', encoding='utf-8') as fobj:
        config_instance().write(fobj)


@pytest.fixture
def get_config_file(config_instance, appdirs):
    """Provide a dynamic config file store under our fake config dir."""
    def generate(**kwargs):
        instance = config_instance(**kwargs)
        conf_path = os.path.join(appdirs.user_config_dir, 'dob.conf')
        with codecs.open(conf_path, 'w', encoding='utf-8') as fobj:
            instance.write(fobj)
        return instance
    return generate


# *** Various config settings

@pytest.fixture
def db_name(request):
    """Return a randomized database name."""
    return fauxfactory.gen_utf8()


@pytest.fixture
def db_user(request):
    """Return a randomized database username."""
    return fauxfactory.gen_utf8()


@pytest.fixture
def db_password(request):
    """Return a randomized database password."""
    return fauxfactory.gen_utf8()


@pytest.fixture(params=(fauxfactory.gen_latin1(), fauxfactory.gen_ipaddr()))
def db_host(request):
    """Return a randomized database username."""
    return request.param


@pytest.fixture
def db_port(request):
    """Return a randomized database port."""
    return str(fauxfactory.gen_integer(min_value=0, max_value=65535))


# ***

@pytest.fixture
def ongoing_fact(controller_with_logging, fact):
    """Fixture tests that ``ongoing fact`` can be saved to data store."""
    fact.end = None
    fact = controller_with_logging.facts.save(fact)
    return fact


# ***

def prepare_controller(config_root):
    controller = Controller()
    controller.wire_configience(config_root=config_root)
    return controller


@pytest.fixture
def test_fact_cls():
    return Fact


@pytest.yield_fixture
def controller(config_root, mocker, test_fact_cls):
    """Provide a pseudo controller instance."""
    controller = prepare_controller(config_root=config_root)
    controller.ctx = mocker.MagicMock()
    controller.configurable = mocker.MagicMock()
    controller.standup_store(fact_cls=test_fact_cls)
    yield controller
    controller.store.cleanup()


@pytest.yield_fixture
def controller_with_logging(config_root, mocker, test_fact_cls):
    """Provide a pseudo controller instance with logging setup."""
    controller = prepare_controller(config_root=config_root)
    controller.ctx = mocker.MagicMock()
    controller.configurable = mocker.MagicMock()
    controller.setup_logging()
    controller.standup_store(fact_cls=test_fact_cls)
    yield controller
    controller.store.cleanup()




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import os

from gettext import gettext as _

from ..termio import click_echo, dob_in_user_exit, highlight_value

from . import ConfigRoot

from .fileboss import (
    default_config_path,
    empty_config_obj,
    load_config_obj,
    write_config_obj
)

__all__ = (
    'ConfigUrable',
)


class ConfigUrable(object):
    """"""

    DOB_CONFIGFILE_ENVKEY = 'DOB_CONFIGFILE'

    def __init__(self):
        super(ConfigUrable, self).__init__()
        self.configfile_path = None
        # The ConfigRoot is a module-level Singleton. Deal.
        self.config_root = ConfigRoot

    # ***

    @property
    def config_path(self):
        return self.configfile_path

    # ***

    def find_all(self, parts):
        return self.config_root.find_all(parts)

    # ***

    def create_config(self, force):
        cfgfile_exists = os.path.exists(self.config_path)
        if cfgfile_exists and not force:
            dob_in_user_exit(_('Config file exists'))

        self.reset_config()
        click_echo(
            _('Initialized default Dob configuration at {}').format(
                highlight_value(self.config_path),
            )
        )

    # ***

    def load_config(self, configfile_path):
        def _load_config():
            self.configfile_path = _resolve_configfile_path(configfile_path)
            cfgfile_exists = os.path.exists(self.config_path)
            config_obj = load_config_obj(self.config_path)
            self.config_root.forget_config_values()
            self.config_root.update_known(config_obj)
            self.cfgfile_exists = cfgfile_exists
            self.cfgfile_sanity = not self.cfgfile_exists or _is_config_like()

        def _resolve_configfile_path(commandline_value):
            if commandline_value is not None:
                return commandline_value

            if ConfigUrable.DOB_CONFIGFILE_ENVKEY in os.environ:
                return os.environ[ConfigUrable.DOB_CONFIGFILE_ENVKEY]

            return default_config_path()

        def _is_config_like():
            # What's a reasonable expectation to see if the config file
            # legitimately exists? Check that the file exists? Or parse it
            # and verify one or more settings therein? Let's do the latter,
            # seems more robust. We can check the `store` settings, seems
            # like the most obvious setting to check. In any case, we do
            # this just to tell the user if they need to create a config;
            # the app will run just fine without a config file, because
            # defaults!
            try:
                self.config_root.asobj.db.orm.value_from_config
                return True
            except AttributeError:
                return False

        _load_config()

    def inject_from_cli(self, *keyvals):
        def _inject_cli_settings():
            for keyval in keyvals:
                process_option(keyval)

        def process_option(keyval):
            key, value = keyval.split('=', 2)
            parts = key.split('.')
            setting = self.config_root.find_setting(parts)
            if setting is None:
                dob_in_user_exit(
                    _('ERROR: Unknown config option: “{}”').format(key)
                )
            setting.value_from_cliarg = value

        return _inject_cli_settings()

    # ***

    def round_out_config(self):
        self.write_config(skip_unset=False)

    # ***

    def reset_config(self):
        config_obj = empty_config_obj(self.config_path)
        # Fill in dict object using Config defaults.
        self.config_root.forget_config_values()
        self.config_root.apply_items(config_obj, use_defaults=True)
        write_config_obj(config_obj)
        self.cfgfile_exists = True  # If anything, we just created it!
        self.cfgfile_sanity = True  # If anything, we just created it!

    # ***

    def write_config(self, skip_unset=False):
        config_obj = empty_config_obj(self.config_path)
        # Fill in dict object using values previously set from config or newly set.
        self.config_root.apply_items(config_obj, skip_unset=skip_unset)
        write_config_obj(config_obj)
        self.cfgfile_exists = True
        self.cfgfile_sanity = True




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Methods to control paging and manage (accumulate) pager output."""

import sys
from functools import update_wrapper

import click_hotoffthehamster as click

from .style import coloring

__all__ = (
    'click_echo',
    'disable_paging',
    'enable_paging',
    'flush_pager',
    'paging',
    'set_paging',
)


# ***

this = sys.modules[__name__]

this.PAGER_ON = False

# (lb): This module-scope global makes me feel somewhat icky.
# MAYBE/2020-02-01: See newly added nark.helpers.singleton.
this.PAGER_CACHE = []


def disable_paging():
    this.PAGER_ON = False


def enable_paging():
    this.PAGER_ON = True


def paging():
    return this.PAGER_ON


def set_paging(new_paging):
    was_paging = this.PAGER_ON
    this.PAGER_ON = new_paging
    return was_paging


# ***

def click_echo(message=None, **kwargs):
    if coloring():
        kwargs['color'] = True
    if not paging():
        click.echo(message, **kwargs)
    else:
        # Collect echoes and show at end, otherwise every call
        # to echo_via_pager results in one pager session, and
        # user has to click 'q' to see each line of output!
        this.PAGER_CACHE.append(message or '')


# ***

def flush_pager(func):
    def flush_echo(*args, **kwargs):
        func(*args, **kwargs)
        if paging() and this.PAGER_CACHE:
            click.echo_via_pager(u'\n'.join(this.PAGER_CACHE))
            this.PAGER_CACHE = []

    return update_wrapper(flush_echo, func)




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Terminal input and output modules."""

# Import herein for convenience references.
# - NOTE: noqa: F401: Disable: 'foo.bar' imported but unused.
from .echoes import echo_block_header, highlight_value  # noqa: F401
from .errors import (  # noqa: F401
    barf_and_exit, dob_in_user_exit, dob_in_user_warning, echo_exit
)
from .paging import click_echo  # noqa: F401
from .style import attr, bg, coloring, colorize, fg  # noqa: F401




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

import sys

import ansi_escape_room

__all__ = (
    'disable_colors',
    'enable_colors',
    'coloring',
    'set_coloring',
    'fg',
    'bg',
    'attr',
    'colorize',
    # Private:
    #  'map_color'
)


# (lb): Retrieve pointer to module object instance, so functions can set.
# - If we did not do this, function scoping would create local variables.
# - I'll admit this feels a little weird, but we need to gait access to the
#   ansi_escape_room package. Using a fake "this", like we do here, is one
#   way to gait access. Another would be to use a module-level dictionary
#   variable that module functions could set. Or we could make a Singleton
#   class. A fourth option would be to move features in this wrapper inside
#   the ansi-escape-room package. (I created this file before creating
#   ansi-escape-room (a fork of another user's package, named "colored"),
#   and it works fine, and it's pretty simple, so keeping it... for now.)
# MAYBE/2020-02-01: What about using nark's new singleton class instead?
this = sys.modules[__name__]

# (lb): We had been enabling colors until disabled, but it seems more
# normal to assume no colors until at least the user's config is read
# and the command line arguments are parsed! The one drawback is that
# now some module- and class-scope strings can no longer be formatted
# when sourced, but must be generated at runtime, after the config is
# loaded.
this.ENABLE_COLORS = False


def disable_colors():
    this.ENABLE_COLORS = False


def enable_colors():
    this.ENABLE_COLORS = True


def coloring():
    return this.ENABLE_COLORS


def set_coloring(new_coloring):
    was_coloring = this.ENABLE_COLORS
    this.ENABLE_COLORS = new_coloring
    return was_coloring


def fg(color):
    if not coloring():
        return ''
    return ansi_escape_room.fg(map_color(color))


def bg(color):
    if not coloring():
        return ''
    return ansi_escape_room.bg(map_color(color))


def attr(color):
    if not coloring():
        return ''
    return ansi_escape_room.attr(map_color(color))


def colorize(text, color, *args):
    if not coloring():
        return text
    more_attrs = ''.join([ansi_escape_room.attr(attr) for attr in args])
    return '{}{}{}{}'.format(
        ansi_escape_room.fg(color),
        more_attrs,
        text,
        ansi_escape_room.attr('reset'),
    )


def map_color(color):
    # FIXME/2018-06-08: (lb): Need a way to easily change palette.
    # Should at least have two profiles, one for black on white; and t'other.
    # Search all uses of fg and bg, and maybe even map attr?
    return color




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""The Controller."""

import inspect
import logging
import os
import sys

from gettext import gettext as _

from nark.control import NarkControl
from nark.helpers import logging as logging_helpers
from nark.items.fact import Fact

# MAYBE/2020-02-01: Have dob set __arg0name__.
from . import __arg0name__
from . import help_newbs as help_strings
from .config.urable import ConfigUrable
from .termio.echoes import click_echo, highlight_value
from .termio.errors import dob_in_user_exit, dob_in_user_warning
from .termio.style import disable_colors, enable_colors

__all__ = (
    'Controller',
)


# ***
# *** [CONTROLLER] NarkControl Controller.
# ***

class Controller(NarkControl):
    """
    A custom controller that adds config handling on top of its regular functionality.
    """

    POST_PROCESSORS = []

    def __init__(self, config=None):
        super(Controller, self).__init__(config)
        self.configurable = None
        self.arg0name = __arg0name__

    @property
    def now(self):
        return self.store.now

    @property
    def data_store_exists_at(self):
        return _(
            'Data store already exists at {}'
        ).format(self.config['db.path'])

    @property
    def data_store_url(self):
        return self.store.db_url

    @property
    def sqlite_db_path(self):
        if self.config['db.engine'] == 'sqlite':
            return self.config['db.path']
        else:
            # (lb): I don't super-like this. It's a weird side effect.
            #   And it's knowledgeable about the CLI command API. Meh.
            dob_in_user_exit(_(
                'Not a SQLite database. Try `{} store url`'
            ).format(self.arg0name))

    @property
    def is_germinated(self):
        if (
            (
                self.configurable.cfgfile_exists
                and self.configurable.cfgfile_sanity
            )
            and self.store_exists
        ):
            return True
        return False

    @property
    def store_exists(self):
        # Check either db.path is set, or all of db.host/port/name/user.
        if self.config['db.engine'] == 'sqlite':
            return os.path.isfile(self.config['db.path'])
        else:
            return bool(self.store.db_url)

    def standup_store(self, fact_cls=Fact):
        self.store.fact_cls = fact_cls
        return super(Controller, self).standup_store()

    def insist_germinated(self, fact_cls=Fact):
        """Assist user if config or database not present."""
        def _insist_germinated():
            if self.is_germinated:
                self.standup_store(fact_cls)
                return
            if not self.configurable.cfgfile_exists and not self.store_exists:
                help_newbie_onboard()
            else:
                berate_user_files_unwell()
            sys.exit(1)

        def help_newbie_onboard():
            message = help_strings.NEWBIE_HELP_ONBOARDING(self.ctx)
            click_echo(inspect.cleandoc(message), err=True)

        def berate_user_files_unwell():
            if not self.configurable.cfgfile_exists:
                oblige_user_create_config()
            if not self.configurable.cfgfile_sanity:
                oblige_user_repair_config()
            if not self.store_exists:
                oblige_user_create_store()

        def oblige_user_create_config():
            cfg_path = self.configurable.config_path
            message = help_strings.NEWBIE_HELP_CREATE_CONFIG(self.ctx, cfg_path)
            click_echo(inspect.cleandoc(message), err=True)

        def oblige_user_repair_config():
            cfg_path = self.configurable.config_path
            message = help_strings.NEWBIE_HELP_REPAIR_CONFIG(self.ctx, cfg_path)
            click_echo(inspect.cleandoc(message), err=True)

        def oblige_user_create_store():
            message = help_strings.NEWBIE_HELP_CREATE_STORE(
                self.ctx,
                db_path=self.config['db.path'],
                val_source=self.config.asobj.db.path.source,
            )
            click_echo(inspect.cleandoc(message), err=True)

        _insist_germinated()

    # ***

    def ensure_config(self, ctx, configfile_path=None, *keyvals):
        if self.configurable is not None:
            return
        self.ctx = ctx
        self.configurable = ConfigUrable()
        self.configurable.load_config(configfile_path)
        self.configurable.inject_from_cli(*keyvals)
        self.wire_configience()

    def create_config(self, force):
        self.configurable.create_config(force=force)
        self.wire_configience()

    def round_out_config(self, force):
        self.configurable.round_out_config()

    def write_config(self, skip_unset=False):
        self.configurable.write_config(skip_unset=skip_unset)

    def wire_configience(self, config_root=None):
        self.config = config_root or self.configurable.config_root
        self.capture_config_lib(self.config)
        self._adjust_log_level()

    def _adjust_log_level(self):
        # *cough*hack!*cough*”
        # Because invoke_without_command, we allow command-less invocations.
        #   For one such invocation -- dob -v -- tell the store not to log.
        # Also tell the store not to log if user did not specify anything,
        #   because we'll show the help/usage (which Click would normally
        #   handle if we had not tampered with invoke_without_command).
        if (
            (len(sys.argv) > 2)
            or (
                (len(sys.argv) == 2)
                and (sys.argv[1] not in ('-v', 'version'))
            )
        ):
            return
        # FIXME/EXPLAIN/2019-01-22: (lb): What about other 2 loggers?
        #   dev.cli_log_level
        #   dev.lib_log_level
        # (lb): Normally I'd prefer the []-lookup vs. attr., e.g., not:
        #   self.config.asobj.dev.sql_log_level.value_from_forced = 'WARNING'
        # because the self.config has non-key-val attributes (like
        # setdefault) so I think for clarity we should lookup via [].
        # Except the []-lookup returns the value, not the keyval object.
        # So here we have to use dotted attribute notation.
        self.config.asobj.dev.sql_log_level.value_from_forced = 'WARNING'

    # ***

    def create_data_store(self, force, fact_cls=Fact):
        skip_standup = self.check_sqlite_store_ready()
        if skip_standup:
            if force:
                self._reset_data_store()
                unlinked_db = True
            else:
                dob_in_user_exit(self.data_store_exists_at)
        self._standup_and_version_store(fact_cls)
        if unlinked_db:
            self._announce_recreated_store()

    def create_config_and_store(self, fact_cls=Fact):
        def _create_config_and_store():
            if not self.is_germinated:
                germinate_config_and_store()
            else:
                exit_already_germinated()

        def germinate_config_and_store():
            create_config_maybe()
            create_store_maybe()

        def create_config_maybe():
            cfg_path = self.configurable.config_path
            if not os.path.exists(cfg_path):
                self.create_config(force=False)
            else:
                click_echo(
                    _('Configuration already exists at {}').format(
                        highlight_value(cfg_path),
                    )
                )

        def create_store_maybe():
            # MEH: (lb): If the engine is not SQLite, this function cannot behave
            # like create_config_maybe, which tells the user if the things exists
            # already, because the storage class, SQLAlchemyStore, blindly calls
            # create_all (in create_storage_tables) without checking if db exists.
            skip_standup = self.check_sqlite_store_ready()
            if skip_standup:
                click_echo(self.data_store_exists_at)
            else:
                self._standup_and_version_store(fact_cls)

        def exit_already_germinated():
            dob_in_user_exit(_(
                'Dob is already setup. Run `{} details` for info.'
            ).format(self.arg0name))

        _create_config_and_store()

    def check_sqlite_store_ready(self):
        if self.config['db.engine'] != 'sqlite':
            return None
        db_path = self.config['db.path']
        if not os.path.isfile(db_path):
            return False
        return True

    def _reset_data_store(self):
        if self.config['db.engine'] != 'sqlite':
            # raise NotImplementedError
            dob_in_user_exit(_(
                'FIXME: Reset non-SQLite data store not supported (yet).'
            ))
        else:
            self.must_unlink_db_path(force=True)

    def must_unlink_db_path(self, *_args, force):
        db_path = self.config['db.path']
        if not os.path.exists(db_path):
            return
        if not os.path.isfile(db_path):
            dob_in_user_exit(_(
                'Data store exists but is not a file, so not overwriting {}'
            ).format(db_path))
        if not force:
            dob_in_user_exit(self.data_store_exists_at)
        os.unlink(db_path)

    def _announce_recreated_store(self):
        click_echo(
            _('Recreated data store at {}')
            .format(highlight_value(self.config['db.path']))
        )

    def _standup_and_version_store(self, fact_cls):
        created_fresh = self.standup_store(fact_cls)
        if created_fresh:
            verb = _('created')
        else:
            verb = _('already ready')
        click_echo(
            _('Dob database {verb} at {url}').format(
                verb=verb, url=highlight_value(self.store.db_url),
            )
        )

    @staticmethod
    def post_processor(func):
        Controller.POST_PROCESSORS.append(func)

    @staticmethod
    def _post_process(controller, fact_facts_or_true, show_plugin_error=None):
        # facts_or_true is one of:
        # - The saved/edited Fact;
        # - a list of Facts;
        # - or True, on upgrade-legacy.
        for handler in Controller.POST_PROCESSORS:
            handler(controller, fact_facts_or_true, show_plugin_error)

    def post_process(self, controller, fact_facts_or_true, show_plugin_error=None):
        Controller._post_process(controller, fact_facts_or_true, show_plugin_error)

    def setup_tty_color(self, use_color):
        if use_color is None:
            use_color = self.config['term.use_color']
        else:
            self.config['term.use_color'] = use_color
        if use_color:
            enable_colors()
        else:
            disable_colors()

    # ***

    def setup_logging(self, verbose=False, verboser=False):
        """Setup logging for the lib_logger as well as client specific logging."""
        self.client_logger = logging.getLogger('dob')
        loggers = self.get_loggers()
        for logger in loggers:
            logger.handlers = []
        # Clear existing Handlers, and set the level.
        # MAYBE: Allow user to specify different levels for different loggers.
        cli_log_level_name = self.config['dev.cli_log_level']
        cli_log_level, warn_name = logging_helpers.resolve_log_level(cli_log_level_name)
        # We can at least allow some simpler optioning from the command args.
        if verbose:
            cli_log_level = min(logging.INFO, cli_log_level)
        if verboser:
            cli_log_level = min(logging.DEBUG, cli_log_level)
        # 2019-01-25 (lb): I have not had any issues for past few weeks, but,
        #   just FYI in case you do, you might need to clear handlers on
        #   lib_logger and sql_logger, e.g.,:
        #        for logger in loggers:
        #            logger.handlers = []
        #            logger.setLevel(cli_log_level)
        self.client_logger.handlers = []
        self.client_logger.setLevel(cli_log_level)

        color = self.config['log.use_color']
        formatter = logging_helpers.formatter_basic(color=color)

        if self.config['log.use_console']:
            console_handler = logging.StreamHandler()
            logging_helpers.setup_handler(console_handler, formatter, *loggers)

        logfile = self.config['log.filepath']
        if logfile:
            file_handler = logging.FileHandler(logfile, encoding='utf-8')
            logging_helpers.setup_handler(file_handler, formatter, *loggers)

        if warn_name:
            self.client_logger.warning(
                _('Unknown Client.cli_log_level specified: {}')
                .format(cli_log_level)
            )

    def get_loggers(self):
        loggers = [
            self.lib_logger,
            self.sql_logger,
            self.client_logger,
        ]
        return loggers

    def bulk_set_log_levels(self, log_level):
        for logger in self.get_loggers():
            logger.setLevel(log_level)

    def disable_logging(self):
        loggers = [
            self.lib_logger,
            self.sql_logger,
            self.client_logger,
        ]
        for logger in loggers:
            logger.handlers = []
            logger.setLevel(logging.NOTSET)

    # ***

    def affirm(self, condition):
        if condition:
            return
        self.client_logger.error(_('Something catastrophic happened!'))
        if not self.config['dev.catch_errors']:
            return
        import traceback
        traceback.print_stack()
        traceback.print_exc()
        self.pdb_set_trace()

    # ***

    def pdb_set_trace(self):
        import pdb
        self.pdb_break_enter()
        pdb.set_trace()
        self.pdb_break_leave()

    def pdb_break_enter(self):
        import subprocess
        # If the developer breaks into code from within the Carousel,
        # i.e., from within the Python Prompt Toolkit library, then
        # pdb terminal echo of stdin back to stdout is broken. You'll
        # see that pdb.stdin and pdb.stdout still match the sys.__stdin__
        # and sys.__stdout__, so that's not the issue -- it's that pdb
        # terminal is in *raw* mode. We can fix this by shelling to stty.
        proc = subprocess.Popen(['stty', '--save'], stdout=subprocess.PIPE)
        (stdout_data, stderr_data) = proc.communicate()
        self.stty_saved = stdout_data.strip()
        # Before breaking, twiddle the terminal away from PPT temporarily.
        subprocess.Popen(['stty', 'sane'])

    def pdb_break_leave(self):
        import subprocess
        # Aha! This is awesome! We can totally recover from an interactive
        # debug session! First, restore the terminal settings (which we
        # reset so the our keystrokes on stdin were echoed back to us)
        # so that sending keys to PPT works again.
        subprocess.Popen(['stty', self.stty_saved])
        # And then the caller, if self.carousel, will redraw the interface
        # (because it has the handle to the application).
        self.client_logger.debug(_("Get on with it!"))

    # ***

    def find_latest_fact(self, restrict=None):
        try:
            return self.facts.find_latest_fact(restrict=restrict)
        except Exception as err:
            # (lb): Unexpected! This could mean more than one ongoing Fact found!
            dob_in_user_warning(str(err))

    def find_oldest_fact(self):
        return self.facts.find_oldest_fact()




# This file exists within 'dob-bright':
#
#   https://github.com/hotoffthehamster/dob-bright
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Hamter CLI Nonbusiness Helper modules."""

import sys

import click_hotoffthehamster as click

from .ascii_art import infection_notice, lifeless
from .paging import click_echo, flush_pager
from .style import colorize

__all__ = (
    'dob_in_user_exit',
    'dob_in_user_warning',
    'barf_and_exit',
    'echo_exit',
)


# ***

def dob_in_user_exit(msg):
    # (lb): I made two similar error-and-exit funcs. See also: barf_and_exit.
    dob_in_user_warning(msg)
    sys.exit(1)


# ***

def dob_in_user_warning(msg):
    click.echo(colorize(msg, 'red_3b'), err=True)


# ***

def barf_and_exit(msg, crude=True):
    # (lb): I made two similar error-and-exit funcs. See also: dob_in_user_exit.
    if crude:
        click_echo()
        click_echo(lifeless().rstrip())
        click_echo(infection_notice().rstrip())
        # click.pause(info='')
    click_echo()
    click_echo(colorize(msg, 'red'))
    sys.exit(1)


# ***

def echo_exit(ctx, message, exitcode=0):
    def _echo_exit(message):
        click_echo(message)
        _flush_pager()
        ctx.exit(exitcode)

    def _flush_pager():
        # To get at the PAGER_CACHE, gotta go through the decorator.
        # So this is quite roundabout.
        @flush_pager
        def __flush_pager():
            pass
        __flush_pager()

    _echo_exit(message)


